from copy import copy

class ParseOptionChange(object):
    def __init__(self, name, obj):
        self._name = name
        self._obj = obj
        self._obj.parse_option_changes = getattr(self._obj, 'parse_option_changes', [])

    def to(self, value):
        self._obj.parse_option_changes.append((self._name, value))
        return self._obj


class ParseOptionBuilder(object):
    def change_parse_option(self, name):
        """
        Change a parse option on the source.

        If there are not yet bytes uploaded, these parse options will be used
        in order to parse the file.

        If there are already bytes uploaded, this will trigger a re-parsing of
        the file, and consequently a new InputSchema will be created. You can call
        `source.latest_input()` to get the newest one.

        Parse options are:
        header_count (int): the number of rows considered a header
        column_header (int): the one based index of row to use to generate the header
        encoding (string): defaults to guessing the encoding, but it can be explicitly set
        column_separator (string): For CSVs, this defaults to ",", and for TSVs "\t", but you can use a custom separator
        quote_char (string): Character used to quote values that should be escaped. Defaults to "\""

        Args:
        ```
            name (string): One of the options above, ie: "column_separator" or "header_count"
        ```

        Returns:
        ```
            change (ParseOptionChange): implements a `.to(value)` function which you call to set the value
        ```

        For our example, assume we have this dataset

        ```
        This is my cool dataset
        A, B, C
        1, 2, 3
        4, 5, 6
        ```

        We want to say that the first 2 rows are headers, and the second of those 2
        rows should be used to make the column header. We would do that like so:

        Examples:
        ```python
            (ok, source) = source\
            .change_parse_option('header_count').to(2)\
            .change_parse_option('column_header').to(2)\
            .run()
        ```

        """
        return ParseOptionChange(name, self)

    def run(self):
        parse_options = copy(self.attributes['parse_options'])
        parse_options.update({key: value for key, value in self.parse_option_changes})

        self.parse_option_changes = []
        return self.update({
            'parse_options': parse_options
        })



from requests.auth import HTTPBasicAuth

class Authorization(object):
    """
    Manages basic authorization for accessing the socrata API.
    This is passed into the `Socrata` object once, which is the entry
    point for all operations.

        auth = Authorization(
            "data.seattle.gov",
            os.environ['SOCRATA_USERNAME'],
            os.environ['SOCRATA_PASSWORD']
        )
        publishing = Socrata(auth)
    """
    def __init__(self, domain, username, password, request_id_prefix = ''):
        self.domain = domain
        self.username = username
        self.password = password

        self.proto = 'https://'
        self.verify = True

        self._request_id_prefix = request_id_prefix

        self.basic = HTTPBasicAuth(self.username, self.password)

    def live_dangerously(self):
        """
        Disable SSL checking. Note that this should *only* be used while developing
        against a local Socrata instance.
        """
        import requests
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        self.verify = False

    def request_id_prefix(self):
        return self._request_id_prefix




import json
import requests
from socrata.http import post, put, delete, get
from socrata.resource import Collection, Resource
from socrata.sources import Source
from socrata.job import Job
import webbrowser

class Revisions(Collection):
    def __init__(self, fourfour, auth):
        self.auth = auth
        self.fourfour = fourfour


    def path(self):
        return 'https://{domain}/api/publishing/v1/revision/{fourfour}'.format(
            domain = self.auth.domain,
            fourfour = self.fourfour
        )

    def _create(self, action_type, metadata, permission):
        body = {
            'metadata': metadata,
            'action': {
                'type': action_type,
                'permission': permission
            }
        }
        return self._subresource(Revision, post(
            self.path(),
            auth = self.auth,
            data = json.dumps(body)
        ))

    def list(self):
        """
        List all the revisions on the view

        Returns:
        ```
            result (bool, dict | list[Revision])
        ```
        """
        return self._subresources(Revision, get(
            self.path(),
            auth = self.auth
        ))


    def create_replace_revision(self, metadata = {}, permission = 'public'):
        """
        Create a revision on the view, which when applied, will replace the data.

        Args:
        ```
            metadata (dict): The metadata to change; these changes will be applied when the revision
                is applied
            permission (string): 'public' or 'private'
        ```
        Returns:
        ```
            result (bool, dict | Revision): The new revision, or an error
        ```
        Examples:
        ```
            >>> view.revisions.create_replace_revision(metadata = {'name': 'new dataset name', 'description': 'updated description'})
        ```
        """
        return self._create('replace', metadata, permission)

    def create_update_revision(self, metadata = {}, permission = 'public'):
        """
        Create a revision on the view, which when applied, will update the data
        rather than replacing it.

        This is an upsert; if there is a rowId defined and you have duplicate ID values,
        those rows will be updated. Otherwise they will be appended.

        Args:
        ```
            metadata (dict): The metadata to change; these changes will be applied when the revision is applied
            permission (string): 'public' or 'private'
        ```

        Returns:
        ```
            result (bool, dict | Revision): The new revision, or an error
        ```

        Examples:
        ```python
            view.revisions.create_update_revision(metadata = {
                'name': 'new dataset name',
                'description': 'updated description'
            })
        ```
        """
        return self._create('update', metadata, permission)

    def create_delete_revision(self, metadata = {}, permission = 'public'):
        """
        Create a revision on the view, which when applied, will delete rows of data.

        This is an upsert; a row id must be set.

        Args:
        ```
            metadata (dict): The metadata to change; these changes will be applied when the revision is applied
            permission (string): 'public' or 'private'
        ```

        Returns:
        ```
            result (bool, dict | Revision): The new revision, or an error
        ```

        Examples:
        ```python
            view.revisions.create_delete_revision(metadata = {
                'name': 'new dataset name',
                'description': 'description'
            })
        ```
        """
        return self._create('delete', metadata, permission)

    @staticmethod
    def new(auth, metadata):
        path = 'https://{domain}/api/publishing/v1/revision'.format(
            domain = auth.domain,
        )

        (ok, response) = result = post(
            path,
            auth = auth,
            data = json.dumps({
                'action': {
                    'type': 'update'
                },
                'metadata': metadata
            })
        )

        if not ok:
            return result

        return (ok, Revision(auth, response))

    def lookup(self, revision_seq):
        """
        Lookup a revision within the view based on the sequence number

        Args:
        ```
            revision_seq (int): The sequence number of the revision to lookup
        ```

        Returns:
        ```
            result (bool, dict | Revision): The Revision resulting from this API call, or an error
        ```
        """
        return self._subresource(Revision, get(
            self.path() + '/' + str(revision_seq),
            auth = self.auth
        ))

    def create_using_config(self, config):
        """
        Create a revision for the given dataset.
        """
        return self._subresource(Revision, post(
            self.path(),
            auth = self.auth,
            data = json.dumps({
                'config': config.attributes['name']
            })
        ))


class Revision(Resource):
    """
    A revision is a change to a dataset
    """

    def create_upload(self, filename):
        """
        Create an upload within this revision

        Args:
        ```
            filename (str): The name of the file to upload
        ```
        Returns:
        ```
            result (bool, dict | Source): The Source created by this API call, or an error
        ```
        """
        return self.create_source({
            'type': 'upload',
            'filename': filename
        })

    def source_from_url(self, url):
        """
        Create a URL source

        Args:
        ```
            url (str): The URL to create the dataset from
        ```
        Returns:
        ```
            result (bool, dict | Source): The Source created by this API call, or an error
        ```
        """
        return self.create_source({
            'type': 'url',
            'url': url
        })

    def source_from_dataset(self):
        """
        Create a dataset source within this revision
        """
        return self.create_source({
            'type': 'view',
            'fourfour': self.view_id()
        })

    def source_from_agent(self, agent_uid, namespace, path):
        """
        Create a source from a connection agent in this revision
        """
        return self.create_source({
          'type': 'connection_agent',
          'agent_uid': agent_uid,
          'namespace': namespace,
          'path': path
        })

    def create_source(self, uri, source_type):
        return self._subresource(Source, post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps({
                'source_type' : source_type
            })
        ))

    def list_sources(self, uri):
        return self._subresources(Source, get(
            self.path(uri),
            auth = self.auth
        ))

    def get_output_schema(self):
        (ok, _) = rev_result = self.show()
        if not ok:
            return rev_result

        output_schema_id = self.attributes['output_schema_id']
        if not output_schema_id:
            return (True, None)

        (ok, sources) = source_result = self.list_sources()
        if not ok:
            return source_result

        output_schemas = [o_s for source in sources for i_s in source.input_schemas for o_s in i_s.output_schemas]
        output_schema = None

        for o_s in output_schemas:
            if o_s.attributes['id'] == output_schema_id:
                output_schema = o_s

        return (True, output_schema)

    def set_output_schema(self, output_schema_id):
        """
        Set the output schema id on the revision. This is what will get applied when
        the revision is applied if no ouput schema is explicitly supplied

        Args:
        ```
            output_schema_id (int): The output schema id
        ```

        Returns:
        ```
            result (bool, dict | Revision): The updated Revision as a result of this API call, or an error
        ```

        Examples:
        ```python
            (ok, revision) = revision.set_output_schema(42)
        ```
        """
        return self.update({'output_schema_id': output_schema_id})


    def discard(self, uri):
        """
        Discard this open revision.

        Returns:
        ```
            result (bool, dict | Revision): The closed Revision or an error
        ```
        """
        return delete(self.path(uri), auth = self.auth)

    def plan(self, uri):
        """
        Return the list of operations this revision will make when it is applied

        Returns:
        ```
            result (bool, dict): The revision plan
        ```
        """
        (ok, resource) = result = get(self.path(uri), auth = self.auth)
        if ok:
            return (ok, resource['resource'])
        return result


    def update(self, uri, body):
        """
        Set the metadata to be applied to the view
        when this revision is applied

        Args:
        ```
            body (dict): The changes to make to this revision
        ```

        Returns:
        ```
            result (bool, dict | Revision): The updated Revision as a result of this API call, or an error
        ```

        Examples:
        ```python
            (ok, revision) = revision.update({
                'metadata': {
                    'name': 'new name',
                    'description': 'new description'
                }
            })
        ```
        """
        return self._mutate(put(
            self.path(uri),
            auth = self.auth,
            data = json.dumps(body),
        ))

    def apply(self, uri, output_schema = None):
        """
        Apply the Revision to the view that it was opened on

        Args:
        ```
            output_schema (OutputSchema): Optional output schema. If your revision includes
                data changes, this should be included. If it is a metadata only revision,
                then you will not have an output schema, and you do not need to pass anything
                here
        ```

        Returns:
        ```
            result (bool, dict | Job): Returns the job that is being run to apply the revision
        ```

        Examples:
        ```
        (ok, job) = revision.apply(output_schema = my_output_schema)
        ```
        """

        if output_schema:
            if not output_schema.attributes['finished_at']:
                (ok, source) = result = output_schema.parent.parent.show()
                if not ok:
                    return result

                source_type = source.attributes['source_type']
                if source_type['type'] == 'view' and not source_type['loaded']:
                    pass
                else:
                    (ok, output_schema) = result = output_schema.wait_for_finish()
                    if not ok:
                        return result

        body = {}

        if output_schema:
            body.update({
                'output_schema_id': output_schema.attributes['id']
            })

        result = self._subresource(Job, put(
            self.path(uri),
            auth = self.auth,
            data = json.dumps(body)
        ))

        self.show() # To mutate ourself and get the job to show up in our attrs

        return result

    def ui_url(self):
        """
        This is the URL to the landing page in the UI for this revision

        Returns:
        ```
            url (str): URL you can paste into a browser to view the revision UI
        ```
        """
        return "https://{domain}/d/{fourfour}/revisions/{seq}".format(
            domain = self.auth.domain,
            fourfour = self.attributes["fourfour"],
            seq = self.attributes["revision_seq"]
        )

    def open_in_browser(self):
        """
        Open this revision in your browser, this will open a window
        """
        webbrowser.open(self.ui_url(), new = 2)

    def view_id(self):
        return self.attributes["fourfour"]



import json
import requests
from socrata.http import post, put, delete, get, gen_headers
from socrata.resource import Collection, Resource
from socrata.sources import Source
from socrata.job import Job
from socrata.revisions import Revisions
import webbrowser


class CoreResource(Resource):
    def _on_response(self, response):
        self.attributes = response

class Views(Collection):
    def path(self):
        return '{proto}{domain}/api/views'.format(
            proto = self.auth.proto,
            domain = self.auth.domain
        )

    def lookup(self, fourfour):
        """
        Lookup the view by ID

        Args:
        ```
            fourfour (str): The view's identifier, ex: abcd-1234
        ```
        Returns:
        ```
            result (bool, Revision | dict): Returns an API Result; the View or an error response
        ```
        """
        return self._subresource(View, get(
            self.path() + '/' + fourfour,
            auth = self.auth
        ))


class View(CoreResource):
    def __init__(self, *args, **kwargs):
        super(CoreResource, self).__init__(*args, **kwargs)
        self.revisions = Revisions(self.attributes['id'], self.auth)

    def delete(self):
        """
        Delete a Socrata view, given its view id
        """
        path = '{proto}{domain}/api/views/{ff}'.format(
            proto = self.auth.proto,
            domain = self.auth.domain,
            ff = self.attributes['id']
        )
        response = requests.delete(
            path,
            headers = gen_headers(),
            auth = self.auth.basic,
            verify = self.auth.verify
        )

        if response.status_code in [200, 201, 202]:
            return (True, {})
        else:
            return (False, response)

    def ui_url(self):
        """
        This is the URL to the landing page in the UI for this view
        """
        return "https://{domain}/d/{fourfour}".format(
            domain = self.auth.domain,
            fourfour = self.attributes["id"]
        )

    def open_in_browser(self):
        """
        Open this revision in your browser, this will open a window
        """
        webbrowser.open(self.ui_url(), new = 2)

    def show(self):
        return super(CoreResource, self).show("/api/views/{ff}".format(
            ff = self.attributes['id']
        ))



import time
import pprint
from socrata.http import noop, get, TimeoutException
import requests

class Collection(object):
    def __init__(self, auth):
        self.auth = auth

    def _subresources(self, klass, result):
        (ok, resources) = result
        if ok:
            return (ok, [klass(self.auth, res, self) for res in resources])
        return result

    def _subresource(self, klass, result):
        (ok, res) = result
        if ok:
            return (ok, klass(self.auth, res, self))
        return result


def parameterize_links(links, id_name, id_val):
    d = {}
    for name, uri in links.items():
        if type(uri) == str:
            d[name] = uri.replace('{%s}' % id_name, str(id_val))
        else:
            d[name] = parameterize_links(uri, id_name, id_val)
    return d

class ChildResourceSpec(object):
    def __init__(
        self,
        parent,
        child_list_name,
        links_namespace,
        response_namespace,
        child_type,
        link_id_attr
    ):
        self._parent = parent
        self._child_list_name = child_list_name
        self._links_namespace = links_namespace
        self._response_namespace = response_namespace
        self._child_type = child_type
        self._link_id_attr = link_id_attr


    def build_children_from(self, parent_response):
        def build_links(child):

            child_link_templates = self._parent.child_ops[self._links_namespace]
            return parameterize_links(
                child_link_templates,
                self._link_id_attr,
                child['id']
            )

        subresources = []
        # This is the actual list of data in the parent response
        response_list = parent_response['resource'][self._response_namespace]
        for child in response_list:
            child_response = {
                'links': build_links(child),
                'resource': child
            }

            # At this point, this looks identical to the response we receive
            # when looking up the child via the API using a GET request,
            # so we can just create the child resource like we would in that case
            result = (True, child_response)

            (_, subresource) = self._parent._subresource(self._child_type, result)
            subresources.append(subresource)

        return self._child_list_name, subresources


class Resource(object):
    def __init__(self, auth, response, parent = None, *args, **kwargs):
        self.auth = auth
        self._on_response(response)
        self.parent = parent

    @classmethod
    def from_uri(cls, auth, uri):
        path = 'https://{domain}{uri}'.format(
            domain = auth.domain,
            uri = uri
        )
        (ok, resp) = result = get(path, auth = auth)
        if ok:
            return (ok, cls(auth, resp))
        return result


    def _clone(self, result):
        (ok, res) = result
        if ok:
            return (ok, self.__class__(self.auth, res, self.parent))
        return result

    def _on_response(self, response):
        self.attributes = response['resource']
        self.links = response['links']
        self._define_operations(self.links)
        self.define_children(response)

    def define_children(self, response):
        for child_list_name, child_list in [
            spec.build_children_from(response)
            for spec in self.child_specs()
        ]:
            setattr(self, child_list_name, child_list)

    def child_specs(self):
        return []

    def path(self, uri):
        return 'https://{domain}{uri}'.format(
            domain = self.auth.domain,
            uri = uri
        )

    def _subresource(self, klass, result, **kwargs):
        (ok, res) = result
        if ok:
            return (ok, klass(self.auth, res, self, **kwargs))
        return result

    def _subresources(self, klass, result):
        (ok, resources) = result
        if ok:
            return (ok, [klass(self.auth, res, self) for res in resources])
        return result

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, pprint.pformat(self.attributes))

    def _define_operations(self, links):
        self_ops = {name: uri for name, uri in links.items() if type(uri) == str}
        child_ops = {name: d for name, d in links.items() if type(d) == dict}

        for name, uri in self_ops.items():
            setattr(self, name, self._dispatch(name, uri))
            setattr(self, '%s_uri' % name, uri)

        setattr(self, 'available_operations', self_ops.keys())
        setattr(self, 'child_ops', child_ops)

    def list_operations(self):
        """
        Get a list of the operations that you can perform on this
        object. These map directly onto what's returned from the API
        in the `links` section of each resource
        """
        return getattr(self, 'available_operations', [])

    # Yes this is kind of terrifying and bad and magic but i can't
    # think of a less obnoxious way to inject the uri
    def _dispatch(self, name, uri):
        og_method_name = '_' + name
        if not hasattr(self, og_method_name):
            og_method = getattr(self, name, self._noop)
            setattr(self, og_method_name, og_method)

        old = getattr(self, og_method_name, self._noop)
        def f(*args, **kwargs):
            return old(uri, *args, **kwargs)
        return f

    def _noop(self, uri, *args, **kwargs):
        raise NotImplementedError("%s is not implemented" % uri)

    def _mutate(self, result):
        (ok, response) = result
        if ok:
            self._on_response(response)
            return (ok, self)
        return result

    # This is just the identity of this resource, so it's easy to abstract
    def show(self, uri):
        return self._mutate(get(
            self.path(uri),
            auth = self.auth
        ))


    def _wait_for_finish(self, is_finished, is_failed, progress, timeout, sleeptime):
        started = time.time()
        while not is_finished(self):
            current = time.time()
            if timeout and (current - started > timeout):
                raise TimeoutException("Timed out after %s seconds waiting for completion for %s" % (timeout, str(self)))
            (ok, me) = self.show()
            progress(self)
            if not ok:
                return (ok, me)
            if is_failed(self):
                return (False, me)
            time.sleep(sleeptime)
        return (True, self)



import time
import json
from copy import deepcopy
from socrata.resource import Resource
from socrata.configs import Config
from socrata.http import noop, put, get, post


class ColumnChange(object):
    def __init__(self, field_name, attribute, output_schema):
        self._field_name = field_name
        self._attribute = attribute
        self._output_schema = output_schema

    def to(self, value):
        def change_fun(col):
            col[self._attribute] = value
            return col

        self._output_schema.column_changes.append((self._field_name, change_fun))

        return self._output_schema

class TransformChange(object):
    def __init__(self, field_name, output_schema):
        self._field_name = field_name
        self._output_schema = output_schema

    def to(self, value):
        def change_fun(col):
            col['transform']['transform_expr'] = value
            return col

        self._output_schema.column_changes.append((self._field_name, change_fun))

        return self._output_schema


class OutputSchema(Resource):
    """
        This is data as transformed from an InputSchema
    """

    def __init__(self, *args, **kwargs):
        super(OutputSchema, self).__init__(*args, **kwargs)
        self.column_changes = []
        self.column_additions = []
        self.column_deletions = []

    def build_config(self, uri, name, data_action):
        """
        Create a new ImportConfig from this OutputSchema. See the API
        docs for what an ImportConfig is and why they're useful
        """
        (ok, res) = result = post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps({
                'name': name,
                'data_action': data_action
            })
        )
        if ok:
            return (ok, Config(self.auth, res, None))
        return result

    def any_failed(self):
        """
        Whether or not any transform in this output schema has failed
        """
        return any([column['transform']['failed_at'] for column in self.attributes['output_columns']])

    def wait_for_finish(self, progress = noop, timeout = None, sleeptime = 1):
        """
        Wait for this dataset to finish transforming and validating. Accepts a progress function
        and a timeout.
        """
        return self._wait_for_finish(
            is_finished = lambda m: m.attributes['completed_at'],
            is_failed = lambda m: m.any_failed(),
            progress = progress,
            timeout = timeout,
            sleeptime = sleeptime
        )

    def _munge_row(self, row):
        row = row['row']
        columns = sorted(self.attributes['output_columns'], key = lambda oc: oc['position'])
        field_names = [oc['field_name'] for oc in columns]
        return {k: v for k, v in zip(field_names, row)}


    def _get_rows(self, uri, offset, limit):
        ok, resp = get(
            self.path(uri),
            params = {'limit': limit, 'offset': offset},
            auth = self.auth
        )

        if not ok:
            return ok, resp

        rows = resp[1:]

        return (ok, [self._munge_row(row) for row in rows])

    def rows(self, uri, offset = 0, limit = 500):
        """
        Get the rows for this OutputSchema. Acceps `offset` and `limit` params
        for paging through the data.
        """
        return self._get_rows(uri, offset, limit)


    def schema_errors(self, uri, offset = 0, limit = 500):
        """
        Get the errors that resulted in transforming into this output schema.
        Accepts `offset` and `limit` params
        """
        return self._get_rows(uri, offset, limit)

    def schema_errors_csv(self):
        """
        Get the errors that results in transforming into this output schema
        as a CSV stream.

        Note that this returns an (ok, Reponse) tuple, where Reponse
        is a python requests Reponse object
        """
        return get(
            self.path(self.schema_errors_uri),
            auth = self.auth,
            headers = {'accept': 'text/csv', 'content-type': 'text/csv'},
            stream = True
        )

    def validate_row_id(self, uri, field_name):
        """
        Set the row id. Note you must call `validate_row_id` before doing this.

        Args:
        ```
            field_name (str): The column to validate as the row id
        ```

        Returns:
        ```
            result (bool, dict): Returns an API Result; where the response says if it can be used as a row id
        ```
        """
        output_column = [oc for oc in self.attributes['output_columns'] if oc['field_name'] == field_name]
        if len(output_column):
            [output_column] = output_column
            transform_id = output_column['transform']['id']

            return get(
                self.path(uri.format(transform_id = transform_id)),
                auth = self.auth
            )
        else:
            return (False, {"reason": "No column with field_name = %s" % field_name})

    def set_row_id(self, field_name = None):
        """
        Set the row id. Note you must call `validate_row_id` before doing this.

        Args:
        ```
            field_name (str): The column to set as the row id
        ```

        Returns:
        ```
            result (bool, OutputSchema | dict): Returns an API Result; the new OutputSchema or an error response
        ```

        Examples:
        ```python
        (ok, new_output_schema) = output.set_row_id('the_id_column')
        ```
        """
        desired_schema = deepcopy(self.attributes['output_columns'])

        for oc in desired_schema:
            oc['is_primary_key'] = (oc['field_name'] == field_name)

        return self.parent.transform({'output_columns': desired_schema})


    def add_column(self, field_name, display_name, transform_expr, description = None):
        """
        Add a column

        Args:
        ```
            field_name (str): The column's field_name, must be unique
            display_name (str): The columns display name
            transform_expr (str): SoQL expression to evaluate and fill the column with data from
            description (str): Optional column description
        ```

        Returns:
        ```
            output_schema (OutputSchema): Returns self for easy chaining
        ```

        Examples:
        ```python
        (ok, new_output_schema) = output
            # Add a new column, which is computed from the `celsius` column
            .add_column('fahrenheit', 'Degrees (Fahrenheit)', '(to_number(`celsius`) * (9 / 5)) + 32', 'the temperature in celsius')
            # Add a new column, which is computed from the `celsius` column
            .add_column('kelvin', 'Degrees (Kelvin)', '(to_number(`celsius`) + 273.15')
            .run()
        ```
        """
        position = len(self.attributes['output_columns']) + len(self.column_additions) - len(self.column_deletions)
        self.column_additions.append({
            'field_name': field_name,
            'display_name': display_name,
            'description': description,
            'position': position,
            'transform': {
                'transform_expr': transform_expr
            }
        })
        return self

    def drop_column(self, field_name):
        """
        Drop the column

        Args:
        ```
            field_name (str): The column to drop
        ```

        Returns:
        ```
            output_schema (OutputSchema): Returns self for easy chaining
        ```

        Examples:
        ```python
            (ok, new_output_schema) = output
                .drop_column('foo')
                .run()
        ```
        """
        self.column_deletions.append(field_name)
        return self

    def change_column_metadata(self, field_name, attribute):
        """
        Change the column metadata. This returns a ColumnChange,
        which implements a `.to` function, which takes the new value to change to

        Args:
        ```
            field_name (str): The column to change
            attribute (str): The attribute of the column to change
        ```

        Returns:
        ```
            change (TransformChange): The transform change, which implements the `.to` function
        ```

        Examples:
        ```python
            (ok, new_output_schema) = output
                # Change the field_name of date to the_date
                .change_column_metadata('date', 'field_name').to('the_date')
                # Change the description of the celsius column
                .change_column_metadata('celsius', 'description').to('the temperature in celsius')
                # Change the display name of the celsius column
                .change_column_metadata('celsius', 'display_name').to('Degrees (Celsius)')
                .run()
        ```
        """
        return ColumnChange(field_name, attribute, self)

    def change_column_transform(self, field_name):
        """
        Change the column transform. This returns a TransformChange,
        which implements a `.to` function, which takes a transform expression.

        Args:
        ```
            field_name (str): The column to change
        ```

        Returns:
        ```
            change (TransformChange): The transform change, which implements the `.to` function
        ```

        Examples:
        ```python
            (ok, new_output_schema) = output
                .change_column_transform('the_date').to('to_fixed_timestamp(`date`)')
                # Make the celsius column all numbers
                .change_column_transform('celsius').to('to_number(`celsius`)')
                # Add a new column, which is computed from the `celsius` column
                .add_column('fahrenheit', 'Degrees (Fahrenheit)', '(to_number(`celsius`) * (9 / 5)) + 32', 'the temperature in celsius')
                .run()
        ```
        """
        return TransformChange(field_name, self)

    def run(self):
        """
        Run all adds, drops, and column changes.


        Returns:
        ```
            result (bool, OutputSchema | dict): Returns an API Result; the new OutputSchema or an error response
        ```

        Examples:
        ```python
            (ok, new_output_schema) = output
                # Change the field_name of date to the_date
                .change_column_metadata('date', 'field_name').to('the_date')
                # Change the description of the celsius column
                .change_column_metadata('celsius', 'description').to('the temperature in celsius')
                # Change the display name of the celsius column
                .change_column_metadata('celsius', 'display_name').to('Degrees (Celsius)')
                # Change the transform of the_date column to to_fixed_timestamp(`date`)
                .change_column_transform('the_date').to('to_fixed_timestamp(`date`)')
                # Make the celsius column all numbers
                .change_column_transform('celsius').to('to_number(`celsius`)')
                # Add a new column, which is computed from the `celsius` column
                .add_column('fahrenheit', 'Degrees (Fahrenheit)', '(to_number(`celsius`) * (9 / 5)) + 32', 'the temperature in celsius')
                .run()
        ```
        """
        columns = [
            c for c in deepcopy(self.attributes['output_columns'])
            if not (c['field_name'] in self.column_deletions)
        ] + self.column_additions

        for (to_change, change_fun) in self.column_changes:
            column = [
                column
                for column in columns
                if to_change == column['field_name']
            ]
            if not column:
                raise ValueError('Column `%s` does not exist' % to_change)
            else:
                [column] = column

            new_column = change_fun(column)

            # if there is an id, this means the column is not brand new
            if 'id' in column:
                new_column['initial_output_column_id'] = column['id']

            def replace_with(c):
                if c == column:
                    return new_column
                return c

            columns = [replace_with(c) for c in columns]

        columns = sorted(columns, key = lambda x: x['position'])

        for p, c in enumerate(columns):
            c['position'] = p + 1

        desired_schema = {
            'output_columns': columns
        }
        self.column_additions = []
        self.column_deletions = []
        self.column_changes = []
        return self.parent.transform(desired_schema)



import json
import requests
from socrata.http import noop, post, get
from socrata.resource import Collection, Resource, ChildResourceSpec
from socrata.output_schema import OutputSchema

class InputSchema(Resource):
    """
    This represents a schema exactly as it appeared in the source
    """
    def transform(self, uri, body):
        """
        Transform this InputSchema into an Output. Returns the
        new OutputSchema. Note that this call is async - the data
        may still be transforming even though the OutputSchema is
        returned. See OutputSchema.wait_for_finish to block until
        the
        """
        return self._subresource(OutputSchema, post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps(body),
        ))

    def latest_output(self, uri):
        """
        Get the latest (most recently created) OutputSchema
        which descends from this InputSchema

        Returns:
            result (bool, OutputSchema | dict): Returns an API Result; the new OutputSchema or an error response
        """
        return self._subresource(OutputSchema, get(
            self.path(uri),
            auth = self.auth,
        ))

    def get_latest_output_schema(self):
        """
        Note that this does not make an API request

        Returns:
            output_schema (OutputSchema): Returns the latest output schema
        """
        return max(self.output_schemas, key = lambda o: o.attributes['id'])

    def child_specs(self):
        return [
            ChildResourceSpec(
                self,
                'output_schemas',
                'output_schema_links',
                'output_schemas',
                OutputSchema,
                'output_schema_id'
            )
        ]



from socrata.resource import Collection
from socrata.sources import Sources
from socrata.configs import Configs
from socrata.views import Views
from socrata.revisions import Revisions
from socrata.operations.configured_job import ConfiguredJob
from socrata.operations.create import Create
from socrata.operations.utils import SocrataException


class Socrata(Collection):
    """
    Top level publishing object.

    All functions making HTTP calls return a result tuple, where the first element in the
    tuple is whether or not the call succeeded, and the second element is the returned
    object if it was a success, or a dictionary containing the error response if the call
    failed. 2xx responses are considered successes. 4xx and 5xx responses are considered failures.
    In the event of a socket hangup, an exception is raised.
    """
    def __init__(self, auth):
        """
        See the `Authorization` class docs for info on how to construct an auth object.
        """
        super(Socrata, self).__init__(auth)
        self.views = Views(auth)
        self.sources = Sources(auth)
        self.configs = Configs(auth)

    def using_config(self, config_name, view):
        """
        Update a dataset, using the configuration that you previously
        created, and saved the name of. Takes the `config_name` parameter
        which uniquely identifies the config, and the `View` object, which can
        be obtained from `socrata.views.lookup('view-id42')`

        Args:
        ```
            config_name (str): The config name
            view (View): The view to update
        ```

        Returns:
        ```
            result (ConfiguredJob): Returns the ConfiguredJob
        ```

        Note:
            Typical usage would be in a context manager block (as demonstrated in the example
            below). In this case, the `ConfiguredJob` is created and immediately launched by way of
            the call to the `ConfiguredJob.csv` method.

        Examples:
        ```
            with open('my-file.csv', 'rb') as my_file:
                (rev, job) = p.using_config(name, view).csv(my_file)
        ```

        """
        (ok, config) = result = self.configs.lookup(config_name)
        if not ok:
            raise SocrataException("Failed to lookup config %s" % config_name, result)
        return ConfiguredJob(self, view=view, config=config)

    def create(self, **kwargs):
        """
        Shortcut to create a dataset. Returns a `Create` object,
        which contains functions which will create a view, upload
        your file, and validate data quality in one step.

        To actually place the validated data into a view, you can call .apply()
        on the revision
        ```
        (revision, output_schema) Socrata(auth).create(
            name = "cool dataset",
            description = "a description"
        ).csv(file)

        (ok, job) = revision.apply(output_schema = output_schema)
        ```

        Args:
        ```
           **kwargs: Arbitrary revision metadata values
        ```

        Returns:
        ```
            result (Revision, OutputSchema): Returns the revision that was created and the
                OutputSchema created from your uploaded file
        ```

        Examples:
        ```python
        Socrata(auth).create(
            name = "cool dataset",
            description = "a description"
        ).csv(open('my-file.csv'))
        ```

        """
        return Create(self, metadata=kwargs)

    def new(self, metadata):
        """
        Create an empty revision, on a view that doesn't exist yet. The
        view will be created for you, and the initial revision will be returned.

        Args:
        ```
            metadata (dict): Metadata to apply to the revision
        ```

        Returns:
        ```
            result (bool, Revision | dict): Returns an API Result; the Revision if it was created
                or an API Error response
        ```

        Examples:
        ```python
            (ok, rev) = Socrata(auth).new({
                'name': 'hi',
                'description': 'foo!',
                'metadata': {
                    'view': 'metadata',
                    'anything': 'is allowed here'

                }
            })
        ```
        """
        return Revisions.new(self.auth, metadata)


__all__ = [
  "authorization",
  "configs",
  "input_schema",
  "output_schema",
  "revisions",
  "sources",
  "upsert_job"
]



import json
import io
import webbrowser
from socrata.http import post, patch, get, noop
from socrata.resource import Resource, Collection, ChildResourceSpec
from socrata.input_schema import InputSchema
from socrata.builders.parse_options import ParseOptionBuilder

class Sources(Collection):
    def path(self):
        return 'https://{domain}/api/publishing/v1/source'.format(
            domain = self.auth.domain
        )

    def lookup(self, source_id):
        """
        Lookup a source

        Args:
        ```
            source_id (int): The id
        ```

        Returns:
        ```
            result (bool, dict | Source): The Source resulting from this API call, or an error
        ```
        """
        return self._subresource(Source, get(
            self.path() + '/' + str(source_id),
            auth = self.auth
        ))

    def create_upload(self, filename):
        """
        Create a new source. Takes a `body` param, which must contain a `filename`
        of the file.

        Args:
        ```
            filename (str): The name of the file you are uploading
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            (ok, upload) = revision.create_upload('foo.csv')
        ```

        """
        return self._subresource(Source, post(
            self.path(),
            auth = self.auth,
            data = json.dumps({
                'source_type' : {
                    'type': 'upload',
                    'filename': filename
                }
            })
        ))

class Source(Resource, ParseOptionBuilder):
    """
    Uploads bytes into the source. Requires content_type argument
    be set correctly for the file handle. It's advised you don't
    use this method directly, instead use one of the csv, xls, xlsx,
    or tsv methods which will correctly set the content_type for you.
    """
    def bytes(self, uri, file_handle, content_type):
        return self._mutate(post(
            self.path(uri),
            auth = self.auth,
            data = file_handle,
            headers = {
                'content-type': content_type
            }
        ))

    def child_specs(self):
        return [
            ChildResourceSpec(
                self,
                'input_schemas',
                'input_schema_links',
                'schemas',
                InputSchema,
                'input_schema_id'
            )
        ]


    def blob(self, file_handle):
        """
        Uploads a Blob dataset. A blob is a file that will not be parsed as a data file,
        ie: an image, video, etc.


        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-blob.jpg', 'rb') as f:
                (ok, upload) = upload.blob(f)
        ```

        """
        source = self
        if self.attributes['parse_options']['parse_source']:
            (ok, cloned) = self.change_parse_option('parse_source').to(False).run()
            assert ok, cloned
            source = cloned

        return source.bytes(file_handle, "application/octet-stream")


    def csv(self, file_handle):
        """
        Upload a CSV, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-file.csv', 'rb') as f:
                (ok, upload) = upload.csv(f)
        ```
        """
        return self.bytes(file_handle, "text/csv")

    def xls(self, file_handle):
        """
        Upload an XLS, returns the new input schema

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-file.xls', 'rb') as f:
                (ok, upload) = upload.xls(f)
        ```
        """
        return self.bytes(file_handle, "application/vnd.ms-excel")

    def xlsx(self, file_handle):
        """
        Upload an XLSX, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-file.xlsx', 'rb') as f:
                (ok, upload) = upload.xlsx(f)
        ```
        """
        return self.bytes(file_handle, "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")

    def tsv(self, file_handle):
        """
        Upload a TSV, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-file.tsv', 'rb') as f:
                (ok, upload) = upload.tsv(f)
        ```
        """
        return self.bytes(file_handle, "text/tab-separated-values")

    def shapefile(self, file_handle):
        """
        Upload a Shapefile, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-shapefile-archive.zip', 'rb') as f:
                (ok, upload) = upload.shapefile(f)
        ```
        """
        return self.bytes(file_handle, "application/zip")

    def kml(self, file_handle):
        """
        Upload a KML file, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-kml-file.kml', 'rb') as f:
                (ok, upload) = upload.kml(f)
        ```
        """
        return self.bytes(file_handle, "application/vnd.google-earth.kml+xml")


    def geojson(self, file_handle):
        """
        Upload a geojson file, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            with open('my-geojson-file.geojson', 'rb') as f:
                (ok, upload) = upload.geojson(f)
        ```
        """
        return self.bytes(file_handle, "application/vnd.geo+json")


    def df(self, dataframe):
        """
        Upload a pandas DataFrame, returns the new source.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            result (bool, Source | dict): Returns an API Result; the new Source or an error response
        ```

        Examples:
        ```python
            import pandas
            df = pandas.read_csv('test/fixtures/simple.csv')
            (ok, upload) = upload.df(df)
        ```
        """
        s = io.StringIO()
        dataframe.to_csv(s, index=False)
        return self.bytes(bytes(s.getvalue().encode()),"text/csv")

    def add_to_revision(self, uri, revision):
        """
        Associate this Source with the given revision.
        """
        return self._clone(patch(
            self.path(uri),
            auth = self.auth,
            data = json.dumps({
                'revision': {
                    'fourfour': revision.attributes['fourfour'],
                    'revision_seq': revision.attributes['revision_seq']
                }
            })
        ))

    def update(self, uri, body):
        return self._clone(post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps(body)
        ))

    def show_input_schema(self, uri, input_schema_id):
        (ok, res) = result = get(
            self.path(uri.format(input_schema_id = input_schema_id)),
            auth = self.auth
        )

        if ok:
            return self._subresource(InputSchema, result)
        return result

    def get_latest_input_schema(self):
        return max(self.input_schemas, key = lambda s: s.attributes['id'])

    def wait_for_finish(self, progress = noop, timeout = None, sleeptime = 1):
        """
        Wait for this dataset to finish transforming and validating. Accepts a progress function
        and a timeout.
        """
        return self._wait_for_finish(
            is_finished = lambda m: m.attributes['finished_at'],
            is_failed = lambda m: m.attributes['failed_at'],
            progress = progress,
            timeout = timeout,
            sleeptime = sleeptime
        )

    def ui_url(self):
        """
        This is the URL to the landing page in the UI for the sources

        Returns:
        ```
            url (str): URL you can paste into a browser to view the source UI
        ```
        """
        if not self.parent:
            raise NotImplementedError("UI for revisionless sources is not implemented (yet). Sorry!")

        revision = self.parent
        return revision.ui_url() +  '/sources/{source_id}/preview'.format(
            source_id = self.attributes['id']
        )

    def open_in_browser(self):
        """
        Open this source in your browser, this will open a window
        """
        webbrowser.open(self.ui_url(), new = 2)



class Operation(object):
    def __init__(self, publish, **kwargs):
        self.publish = publish
        self.properties = kwargs

    def csv(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload a file of
        type CSV and wait for validation to complete. Returns
        an `OutputSchema` which, when applied, will by applied
        to the view.
        """
        return self.run(file, lambda upload: upload.csv(file), *args, **kwargs)

    def xls(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload a file of
        type XLS and wait for validation to complete. Returns
        an `OutputSchema` which, when applied, will by applied
        to the view.
        """
        return self.run(file, lambda upload: upload.xls(file), *args, **kwargs)

    def xlsx(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload a file of
        type XLSX and wait for validation to complete. Returns
        an `OutputSchema` which, when applied, will by applied
        to the view.
        """
        return self.run(file, lambda upload: upload.xlsx(file), *args, **kwargs)

    def tsv(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload a file of
        type TSV and wait for validation to complete. Returns
        an `OutputSchema` which, when applied, will by applied
        to the view.
        """
        return self.run(file, lambda upload: upload.tsv(file), *args, **kwargs)

    def df(self, dataframe, filename="Dataframe", *args, **kwargs):
        """
        Create a revision on that view, then upload the contents
        of a pandas dataframe and wait for validation to complete.
        Returns an `OutputSchema` which, when applied, will by
        applied to the view.
        """
        return self.run(dataframe, lambda upload: upload.df(dataframe), filename, *args, **kwargs)


    def shapefile(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload the contents
        of the shapefile and wait for validation to complete.
        Returns an `OutputSchema` which, when applied, will by
        applied to the view.
        """
        return self.run(file, lambda upload: upload.shapefile(file), *args, **kwargs)


    def kml(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload the contents
        of the kml and wait for validation to complete.
        Returns an `OutputSchema` which, when applied, will by
        applied to the view.
        """
        return self.run(file, lambda upload: upload.kml(file), *args, **kwargs)

    def geojson(self, file, *args, **kwargs):
        """
        Create a revision on that view, then upload the contents
        of the geojson and wait for validation to complete.
        Returns an `OutputSchema` which, when applied, will by
        applied to the view.
        """
        return self.run(file, lambda upload: upload.geojson(file), *args, **kwargs)



from socrata.http import noop
from socrata.resource import Resource

class Job(Resource):
    def is_complete(self):
        """
        Has this job finished or failed
        """
        status = self.attributes['status']
        return (status == 'failure' or status == 'successful')

    def wait_for_finish(self, progress = noop, timeout = None, sleeptime = 1):
        """
        Wait for this dataset to finish transforming and validating. Accepts a progress function
        and a timeout.
        """
        return self._wait_for_finish(
            is_finished = lambda m: m.attributes['finished_at'],
            is_failed = lambda m: m.attributes['status'] == 'failure',
            progress = progress,
            timeout = timeout,
            sleeptime = sleeptime
        )



from socrata import Socrata
from socrata.authorization import Authorization
from socrata.revisions import Revisions, Revision
from socrata.sources import Sources, Source
from socrata.input_schema import InputSchema
from socrata.output_schema import OutputSchema
from socrata.job import Job
from socrata.configs import Configs, Config
import inspect
import sys
import re

classes = [
  Socrata,
  Authorization,
  Revisions, Revision,
  Sources, Source,
  Configs, Config,
  InputSchema,
  OutputSchema,
  Job
]

def arg_spec_str(thing):
    spec = inspect.getargspec(thing)
    args = spec.args

    # Hacks for closing over self and uri
    if len(args) and args[0] == 'self':
        args = args[1:]
    if len(args) and args[0] == 'uri':
        args = args[1:]

    if not len(args):
        return ''

    arg_str = '\n    Args: ' + ', '.join(args)

    default_str = ''
    if spec.defaults and any([d != None for d in spec.defaults]):
        defaults = list(spec.defaults)
        default_str = '\n    Defaults: ' + ', '.join([
            name + '=' + str(default)
            for (name, default) in zip(args, defaults) if default != None
        ])

    return '```\nArgSpec{arg_str}{default_str}\n```'.format(
        arg_str = arg_str,
        default_str = default_str
    )

def link_to(thing):
    [_, rel] = inspect.getsourcefile(thing).split('socrata-py')
    (_lines, line_num) = inspect.getsourcelines(thing)

    return 'https://github.com/socrata/socrata-py/blob/master/{rel}#L{line_num}'.format(
        rel = rel,
        line_num = line_num
    )

def format_doc(doc):
    return doc



def class_lines(klass):
    doc = inspect.getdoc(klass)
    return [
        '',
        '### [%s](%s)' % (klass.__name__, link_to(klass)),
        arg_spec_str(klass),
        '',
        format_doc(doc or '')
    ]

def func_lines(funcname, func):
    doc = inspect.getdoc(func)
    if doc:
        return [
            '',
            '#### [%s](%s)' % (funcname, link_to(func)),
            arg_spec_str(func),
            '',
            format_doc(doc)
        ]
    return []


lines = ['<!-- doc -->', '# Library Docs']

for klass in classes:
    lines += class_lines(klass)

    funcs = inspect.getmembers(klass, predicate=inspect.isfunction)

    for (funcname, func) in funcs:
        if not funcname.startswith('_'):
            lines += func_lines(funcname, func)

lines.append('<!-- docstop -->')


with open('README.md', 'r') as f:
    readme = f.read()
    docs = '\n'.join(lines)
    p = re.compile('<!-- doc -->.*<!-- docstop -->', re.DOTALL)
    updated_readme = p.sub(docs, readme)


with open('README.md', 'w') as f:
    f.write(updated_readme)





class SocrataException(Exception):
    def __init__(self, message, response):
        super(Exception, self).__init__(message + ':\n' + str(response))
        self.response = response

def get_filename(data, filename):
    if (filename is None) and getattr(data, 'name', False):
        filename = data.name
    elif filename is None:
        raise SocrataException("When creating without a file handle, you must include the filename as the second argument")
    return filename



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))


setup(
    name='socrata-py',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.4.16',

    description="SDK For publishing to Socrata",
    long_description="""
        SDK for using the data upload, transformation, and validation
        APIs provided by Socrata.
    """,

    # The project's main homepage.
    url='https://github.com/socrata/socrata-py',

    # Author details
    author='Chris Duranti',
    author_email='chris.duranti@socrata.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='socrata open-data not-publishing',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['bin', 'test', 'examples']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["requests"],
)



import json
import requests
from socrata.http import get, post, patch, delete
from socrata.resource import Collection, Resource
from socrata.builders.parse_options import ParseOptionBuilder


class Configs(Collection):
    def path(self):
        return 'https://{domain}/api/publishing/v1/config'.format(
            domain = self.auth.domain
        )

    def create(self, name, data_action, parse_options = None, columns = None):
        """
        Create a new ImportConfig. See http://docs.socratapublishing.apiary.io/
        ImportConfig section for what is supported in `data_action`, `parse_options`,
        and `columns`.
        """
        return self._subresource(Config, post(
            self.path(),
            auth = self.auth,
            data = json.dumps({
                'name': name,
                'data_action': data_action,
                'parse_options': parse_options,
                'columns': columns
            })
        ))

    # ~~ danger ~~ This URL is hardcoded
    def lookup(self, name):
        """
        Obtain a single ImportConfig by name
        """
        return self._subresource(Config, get(
            self.path() + '/' + name,
            auth = self.auth
        ))

    def list(self):
        """
        List all the ImportConfigs on this domain
        """
        return self._subresources(Config, get(
            self.path(),
            auth = self.auth
        ))

class Config(Resource, ParseOptionBuilder):
    def delete(self, uri):
        """
        Delete this ImportConfig. Note that this cannot be undone.
        """
        return delete(self.path(uri), auth = self.auth)

    def update(self, uri, body):
        """
        Mutate this ImportConfig in place. Subsequent revisions opened against this
        ImportConfig will take on its new value.
        """
        return self._mutate(patch(
            self.path(uri),
            auth = self.auth,
            data = json.dumps(body)
        ))

    def create_revision(self, uri, fourfour):
        """
        Create a new Revision in the context of this ImportConfig.
        Sources that happen in this Revision will take on the values
        in this Config.
        """
        # Because of circular dependencies ;_;
        from socrata.revisions import Revision

        (ok, res) = result = post(
            self.path(uri).format(fourfour = fourfour),
            auth = self.auth
        )
        if not ok:
            return result
        return (ok, Revision(self.auth, res))



from socrata.operations.utils import get_filename, SocrataException
from socrata.operations.operation import Operation

class ConfiguredJob(Operation):
    def run(self, data, put_bytes, filename = None):
        filename = get_filename(data, filename)

        (ok, rev) = self.properties['view'].revisions.create_using_config(
            self.properties['config']
        )
        if not ok:
            raise SocrataException("Failed to create the revision", rev)

        (ok, source) = rev.create_upload(filename)
        if not ok:
            raise SocrataException("Failed to create the upload", source)

        (ok, source) = put_bytes(source)
        if not ok:
            raise SocrataException("Failed to upload the file", source)

        output_schema = source.get_latest_input_schema().get_latest_output_schema()

        (ok, output_schema) = output_schema.wait_for_finish()
        if not ok:
            raise SocrataException("The dataset failed to validate", output_schema)

        (ok, job) = rev.apply(output_schema = output_schema)
        if not ok:
            raise SocrataException("Failed to apply the change", job)
        return (rev, job)



import json
import os
import requests
import binascii
import logging
log = logging.getLogger(__name__)

class TimeoutException(Exception):
    pass

def noop(*args, **kwargs):
    pass

def generate_request_id(length = 32):
    return binascii.hexlify(os.urandom(int(length/2))).decode('utf8')

def gen_headers(extra = {}, auth = None):

    request_id_prefix = ''
    if auth:
        request_id_prefix = auth.request_id_prefix()
    d =  {
        'user-agent': 'publish-py',
        'content-type': 'application/json',
        'accept': 'application/json',
        'x-socrata-requestid': request_id_prefix + generate_request_id(32 - len(request_id_prefix))
    }
    d.update(extra)
    return d


def prepare(headers, auth):
    all_headers = gen_headers(headers, auth)
    return all_headers, all_headers['x-socrata-requestid']

def is_json(response):
    return 'application/json' in response.headers['Content-Type']

def respond(response, request_id = None):
    try:
        if response.status_code in [200, 201, 202]:
            if is_json(response):
                return (True, response.json())
            else:
                return (True, response)
        else:
            log.warning("Request failed with %s, request_id was %s", response.status_code, request_id)
            if is_json(response):
                return (False, response.json())
            else:
                return (False, response)
    except Exception: # json.decoder.JSONDecodeError isn't always a thing???? WHY PYTHON
        log.error("Request raised an exception, request_id was %s", request_id)
        return (False, {'error': 'json', 'content': response.content})

def post(path, auth = None, data = None, headers = {}):
    (headers, request_id) = prepare(headers, auth)
    log.info('POST %s %s', path, request_id)
    return respond(requests.post(
        path,
        headers = headers,
        auth = auth.basic,
        verify = auth.verify,
        data = data
    ), request_id = request_id)



def put(path, auth = None, data = None, headers = {}):
    (headers, request_id) = prepare(headers, auth)
    log.info('PUT %s %s', path, request_id)
    return respond(requests.put(
        path,
        headers = headers,
        auth = auth.basic,
        verify = auth.verify,
        data = data
    ), request_id = request_id)


def patch(path, auth = None, data = None, headers = {}):
    (headers, request_id) = prepare(headers, auth)
    log.info('PATCH %s %s', path, request_id)
    return respond(requests.patch(
        path,
        headers = headers,
        auth = auth.basic,
        verify = auth.verify,
        data = data
    ), request_id = request_id)

def get(path, auth = None, params = {}, headers = {}, **kwargs):
    (headers, request_id) = prepare(headers, auth)
    log.info('GET %s %s', path, request_id)
    return respond(requests.get(
        path,
        params = params,
        headers = headers,
        auth = auth.basic,
        verify = auth.verify,
        **kwargs
    ), request_id = request_id)

def delete(path, auth = None, headers = {}):
    (headers, request_id) = prepare(headers, auth)
    log.info('DELETE %s %s', path, request_id)
    return respond(requests.delete(
        path,
        headers = headers,
        auth = auth.basic,
        verify = auth.verify
    ), request_id = request_id)
